
#include "postgres.h"
#include "catalog/pg_type.h" /* For xxxOID constants */
#include "postgres_fe.h"
#include "libpq-fe.h"
#include "libpq-int.h"
#include "utils/inet.h" /* for PGSQL_AF_INET and PGSQL_AF_INET6 */

#ifndef _PQMAXPARAM
#define _PQMAXPARAM 100
#elif _PQMAXPARAM < 8 || _PQMAXPARAM > 9999
#error _PGMAXPARAM must be between 8 and 9999 (inclusive)
#endif

/* Allows user to pass direct text or bytea pointers for param values.
 * Normally, a param value is copied into private library memory.  These
 * 2 types can be very large values and it is very likely the user had
 * to allocate their own memory.  No need for this library to allocate
 * that memory again.  PQput remaps these to TEXTOID and BYTEAOID.
 */
#define TEXTPTROID  3000000
#define BYTEAPTROID 3000001

/*
 * Each param value requires an oid, ptr and 2 ints (oid, value, length, format).
 * This makes the maximum size of a param 20 bytes. A stack size of 4k
 * would allow for 204 query params.	If more params are needed,
 * heap memory is used. Needing more than 204 param values in a query
 * is very rare, although possible.
 */
#define PARAM_STACKSIZE 4096

/*
 * Declares some variables for the param value arrays:
 * oid, values, lengths, formats. These values are assigned by
 * paramArrayAssign().
 */
#define PARAM_ARRAY_DECL \
	char _stackbuffer[PARAM_STACKSIZE]; \
	char *buf = _stackbuffer; \
	Oid *oids = NULL; \
	char **vals = NULL; \
	int *lens = NULL; \
	int *fmts = NULL

/*
 * On the very rare occasion that the number of params is too large for
 * the _stackbuffer, we have to free our allocated buffer.
 */
#define PARAM_ARRAY_FREE do{ \
	if(buf != _stackbuffer) \
		free(buf); \
}while(0)

#define GETPARAM(_conn) &(_conn)->param

/* puts points from either a PGpath or PGpolygon. If its a path,
 * closed indicates whether the path is open or closed. Closed is
 * ignored when a polygon.
 */
static int putptarray(PGconn *conn, int npts, PGpoint *pts, Oid oid, int closed);

/*
 * Assigns param values to param arrays, for use with postgres
 * parameter API. When needed, malloc/free is used for these arrays.
 * Returns 1 on success and 0 on error.
 *
 * This function used to be a macro, but it grow over time ... long macros
 * tend to be annoying.
 */
static int paramArrayAssign(
	PGconn *conn,
	char **buf, /* in=stack, out=stack|heap */
	Oid **oids,
	char ***vals,
	int **lens,
	int **fmts);

static int putv(
	PGconn *conn,
	const char *cmdfmt,
	char **out,
	int outl,
	va_list args);

void pqResetParam(PGconn *conn)
{
	if(conn)
		conn->param.vcnt = 0;
}

void pqClearParam(PGconn *conn)
{
	int i;
	PGparam *param;

	if(!conn)
		return;

	param = GETPARAM(conn);

	for(i=0; i < param->vmax; i++)
		if(param->vals[i].ptr)
			free(param->vals[i].ptr);

	if(param->vals)
		free(param->vals);

	param->vcnt = 0;
	param->vmax = 0;
	param->vals = NULL;
}

int PQparamExec(
	PGconn *conn,
	const char *command,
	int resultFormat,
	PGresult **resultp)
{
	PGresult *res;
	PARAM_ARRAY_DECL;

	if(resultp)
		*resultp = NULL;

	if(!conn)
		return 0;

	if(!paramArrayAssign(conn, &buf, &oids, &vals, &lens, &fmts))
	{
		pqResetParam(conn);
		return 0;
	}

	res = PQexecParams(
		conn,
		command,
		conn->param.vcnt,
		(const Oid *)oids,
		(const char * const *)vals,
		(const int *)lens,
		(const int *)fmts,
		resultFormat);

	PARAM_ARRAY_FREE;
	pqResetParam(conn);

	if(!res)
		return 0;

	/* If caller doesn't want result, clear it. */
	if(!resultp)
		PQclear(res);
	else
		*resultp = res;

	return 1;
}

int PQparamExecf(
	PGconn *conn,
	int resultFormat,
	PGresult **resultp,
	const char *cmdfmt,
	...)
{
	int r;
	va_list args;
	char _buffer[4096];
	char *cmd = _buffer;

	va_start(args, cmdfmt);
	r = putv(conn, cmdfmt, &cmd, (int)sizeof(_buffer), args);
	va_end(args);

	if(r)
		r = PQparamExec(conn, cmd, resultFormat, resultp);
	else
		pqResetParam(conn);

	if(cmd != _buffer)
		free(cmd);
	return r;
}

int PQparamExecPrepared(
	PGconn *conn,
	const char *stmtName,
	int resultFormat,
	PGresult **resultp)
{
	PGresult *res;
	PARAM_ARRAY_DECL;

	if(resultp)
		*resultp = NULL;

	if(!conn)
		return 0;

	if(!paramArrayAssign(conn, &buf, &oids, &vals, &lens, &fmts))
	{
		pqResetParam(conn);
		return 0;
	}

	res = PQexecPrepared(
		conn,
		stmtName,
		conn->param.vcnt,
		(const char * const *)vals,
		(const int *)lens,
		(const int *)fmts,
		resultFormat);

	PARAM_ARRAY_FREE;
	pqResetParam(conn);

	if(!res)
		return 0;

	/* If caller doesn't want result, clear it. */
	if(!resultp)
		PQclear(res);
	else
		*resultp = res;

	return 1;
}

int PQparamSend(PGconn *conn, const char *command, int resultFormat)
{
	int r;
	PARAM_ARRAY_DECL;

	if(!conn)
		return 0;

	if(!paramArrayAssign(conn, &buf, &oids, &vals, &lens, &fmts))
	{
		pqResetParam(conn);
		return 0;
	}

	r = PQsendQueryParams(
		conn,
		command,
		conn->param.vcnt,
		(const Oid *)oids,
		(const char * const *)vals,
		(const int *)lens,
		(const int *)fmts,
		resultFormat);

	PARAM_ARRAY_FREE;
	pqResetParam(conn);
	return r;
}

int PQparamSendf(
	PGconn *conn,
	int resultFormat,
	const char *cmdfmt,
	...)
{
	int r;
	va_list args;
	char _buffer[4096];
	char *cmd = _buffer;

	va_start(args, cmdfmt);
	r = putv(conn, cmdfmt, &cmd, (int)sizeof(_buffer), args);
	va_end(args);

	if(r)
		r = PQparamSend(conn, cmd, resultFormat);
	else
		pqResetParam(conn);

	if(cmd != _buffer)
		free(cmd);
	return r;
}

int PQparamSendPrepared(PGconn *conn, const char *stmtName, int resultFormat)
{
	int r;
	PARAM_ARRAY_DECL;

	if(!conn)
		return 0;

	if(!paramArrayAssign(conn, &buf, &oids, &vals, &lens, &fmts))
	{
		pqResetParam(conn);
		return 0;
	}

	r = PQsendQueryPrepared(
		conn,
		stmtName,
		conn->param.vcnt,
		(const char * const *)vals,
		(const int *)lens,
		(const int *)fmts,
		resultFormat);

	PARAM_ARRAY_FREE;
	pqResetParam(conn);
	return r;
}

int PQputchar(PGconn *conn, char ch)
{
	return PQput(conn, (const void *)&ch, 1, CHAROID, 1);
}

int PQputbool(PGconn *conn, int b)
{
	char c = b ? 't' : 'f';
	return PQput(conn, (const void *)&c, 1, BOOLOID, 1);
}

#if 0
int PQputmoney(PGconn *conn, double money)
{
	/* money is changing to 64-bit. What length do we specify? */
}
#endif

int PQputint2(PGconn *conn, short i2)
{
	i2 = (short)htons(i2);
	return PQput(conn, (const void *)&i2, 2, INT2OID, 1);
}

int PQputint4(PGconn *conn, int i4)
{
	i4 = (int)htonl(i4);
	return PQput(conn, (const void *)&i4, 4, INT4OID, 1);
}

int PQputfloat4(PGconn *conn, float f4)
{
	void *voidp = &f4;
	int i = (int)htonl(*(int *)voidp);
	return PQput(conn, (const void *)&i, 4, FLOAT4OID, 1);
}

int PQputint8v(PGconn *conn, size_t valsize, ...)
{
	va_list args;
	size_t val[2];
	unsigned int buf[2];

	va_start(args, valsize);
	memset(val, 0, sizeof(val));

	switch(valsize)
	{
		case 1:
			val[0] = (unsigned char)va_arg(args, int);
			break;

		case 2:
			val[0] = (unsigned short)va_arg(args, int);
			break;

		case 4:
			val[0] = (unsigned int)va_arg(args, int);
			break;

		case 8:
			val[0] = (size_t)va_arg(args, size_t);
			if(sizeof(size_t) == 4)
				val[1] = (size_t)va_arg(args, size_t);
			break;

		default:
			printfPQExpBuffer(&conn->errorMessage, "Invalid integer size: %zu",
				valsize);
			return 0;
	}

	return PQput(conn, pqSwapi8(buf, val, 1), 8, INT8OID, 1);
}

int PQputfloat8(PGconn *conn, double f8)
{
	unsigned int buf[2];
	return PQput(conn, pqSwapi8(buf, &f8, 1), 8, FLOAT8OID, 1);
}

int PQputtext(PGconn *conn, const char *text)
{
	if(!text)
		return PQputnull(conn);
	return PQput(conn, text, (int)strlen(text) + 1, TEXTOID, 0);
}

int PQputtextptr(PGconn *conn, const char *text)
{
	if(!text)
		return PQputnull(conn);
	return PQput(conn, text, (int)strlen(text) + 1, TEXTPTROID, 0);
}

int PQputbytea(PGconn *conn, const void *bytea, int len)
{
	if(!bytea)
		return PQputnull(conn);
	return PQput(conn, bytea, len, BYTEAOID, 1);
}

int PQputbyteaptr(PGconn *conn, const void *bytea, int len)
{
	if(!bytea)
		return PQputnull(conn);
	return PQput(conn, bytea, len, BYTEAPTROID, 1);
}

int PQputpoint(PGconn *conn, PGpoint *pt)
{
	unsigned int buf[4];

	if(!pt)
		return PQputnull(conn);

	pqSwapi8(buf,	    &pt->x, 1);
	pqSwapi8(buf + 2, &pt->y, 1);
	return PQput(conn, buf, 16, POINTOID, 1);
}

int PQputlseg(PGconn *conn, PGlseg *lseg)
{
	unsigned int buf[8];

	if(!lseg)
		return PQputnull(conn);

	pqSwapi8(buf,     &lseg->pts[0].x, 1);
	pqSwapi8(buf + 2, &lseg->pts[0].y, 1);
	pqSwapi8(buf + 4, &lseg->pts[1].x, 1);
	pqSwapi8(buf + 6, &lseg->pts[1].y, 1);
	return PQput(conn, buf, 32, LSEGOID, 1);
}

int PQputbox(PGconn *conn, PGbox *box)
{
	unsigned int buf[8];

	if(!box)
		return PQputnull(conn);

	pqSwapi8(buf,	    &box->high.x, 1);
	pqSwapi8(buf + 2, &box->high.y, 1);
	pqSwapi8(buf + 4, &box->low.x,  1);
	pqSwapi8(buf + 6, &box->low.y,  1);
	return PQput(conn, buf, 32, BOXOID, 1);
}

int PQputcircle(PGconn *conn, PGcircle *circle)
{
	unsigned int buf[6];

	if(!circle)
		return PQputnull(conn);

	pqSwapi8(buf,	    &circle->center.x, 1);
	pqSwapi8(buf + 2, &circle->center.y, 1);
	pqSwapi8(buf + 4, &circle->radius,   1);
	return PQput(conn, buf, 24, CIRCLEOID, 1);
}

int PQputpath(PGconn *conn, PGpath *path)
{
	if(!path)
		return PQputnull(conn);
	return putptarray(conn, path->npts, path->pts, PATHOID, path->closed ? 1 : 0);
}

int PQputpolygon(PGconn *conn, PGpolygon *polygon)
{
	if(!polygon)
		return PQputnull(conn);
	return putptarray(conn, polygon->npts, polygon->pts, POLYGONOID, -1);
}

int PQputinet(PGconn *conn, PGinet *inet)
{
	unsigned char buf[80];
	unsigned char *b = buf;

	if(!inet)
		return PQputnull(conn);

	if(inet->sa_family == AF_INET)
	{
		struct sockaddr_in *sa = (struct sockaddr_in *)inet->sa_buf;
		*b++ = (unsigned char)PGSQL_AF_INET;
		*b++ = (unsigned char)inet->mask;
		*b++ = (unsigned char)(inet->is_cidr ? 1 : 0);
		*b++ = (unsigned char)4;
		memcpy(b, &sa->sin_addr, 4);
		b += 4;
	}
	else if(inet->sa_family == AF_INET6)
	{
		struct sockaddr_in6 *sa = (struct sockaddr_in6 *)inet->sa_buf;
		*b++ = (unsigned char)PGSQL_AF_INET6;
		*b++ = (unsigned char)inet->mask;
		*b++ = (unsigned char)(inet->is_cidr ? 1 : 0);
		*b++ = (unsigned char)16;
		memcpy(b, &sa->sin6_addr, 16);
		b += 16;
	}
	else
	{
		printfPQExpBuffer(&conn->errorMessage, "Invalid inet family");
		return 0;
	}

	return PQput(conn, buf, (int)(b-buf), inet->is_cidr ? CIDROID : INETOID, 1);
}

int PQputinetstr(PGconn *conn, const char *inetstr)
{
	if(!inetstr)
		return PQputnull(conn);
	return PQput(conn, inetstr, strlen(inetstr)+1, INETOID, 0);
}

int PQputmacaddr(PGconn *conn, PGmacaddr *mac)
{
	unsigned char buf[6];

	if(!mac)
		return PQputnull(conn);

	buf[0] = (unsigned char)mac->a;
	buf[1] = (unsigned char)mac->b;
	buf[2] = (unsigned char)mac->c;
	buf[3] = (unsigned char)mac->d;
	buf[4] = (unsigned char)mac->e;
	buf[5] = (unsigned char)mac->f;
	return PQput(conn, buf, 6, MACADDROID, 1);
}

int PQput(
	PGconn *conn,
	const void *data,
	int datal,
	Oid oid,
	int format)
{
	PGvalue *v;
	PGparam *param;

	if(!conn)
		return 0;

	if(format == 1)
	{
		int l = 0;

		switch(oid)
		{
			case BOOLOID:
			case CHAROID:
				l = 1;
				break;

			case INT2OID:
				l = 2;
				break;

			case FLOAT4OID:
			case INT4OID:
				l = 4;
				break;

			case FLOAT8OID:
			case INT8OID:
				l = 8;
				break;

			/*case MONEYOID: 4|8 bytes, depends on server version*/
		}

		if(l && l != datal)
		{
			printfPQExpBuffer(&conn->errorMessage,
				libpq_gettext("Invalid type length: typlen=%d != inlen=%d"), l, datal);
			pqResetParam(conn);
			return 0;
		}
	}

	param = GETPARAM(conn);

	/* need to grow param vals array */
	if(param->vcnt == param->vmax)
	{
		PGvalue *vals;
		int vmax = param->vmax ? (param->vmax * 3) / 2 : 16;

		if(!(vals = (PGvalue *)realloc(param->vals, sizeof(PGvalue) * vmax)))
		{
			printfPQExpBuffer(&conn->errorMessage,
				libpq_gettext("cannot allocate param value array"));
			pqResetParam(conn);
			return 0;
		}

		/* zero out the new array elements */
		memset(vals + param->vcnt, 0, (vmax - param->vcnt) * sizeof(PGvalue));
		param->vmax = vmax;
		param->vals = vals;
	}

	/* grab next param value */
	v = &param->vals[param->vcnt];

	if(!data)
		datal = 0;

	if(oid == BYTEAPTROID || oid == TEXTPTROID)
	{
		v->data = (char *)data; /* wants direct pointer */
		oid = oid==BYTEAPTROID ? BYTEAOID : TEXTOID;
	}
	else
	{
		/* need more mem for param value ptr */
		if(v->ptrl < datal)
		{
			char *ptr = (char *)realloc(v->ptr, datal);

			if(!ptr)
			{
				printfPQExpBuffer(&conn->errorMessage,
					libpq_gettext("cannot allocate param value data pointer"));
				pqResetParam(conn);
				return 0;
			}

			v->ptrl = datal;
			v->ptr = ptr;
		}

		/* assign the value, length and format to the param value */
		v->data = data ? (char *)memcpy(v->ptr, data, datal) : NULL;
	}

	v->datal	= datal;
	v->oid		= oid;
	v->format = format;
	param->vcnt++;
	return 1;
}

int PQputf(PGconn *conn, const char *format, ...)
{
	int r;
	va_list args;

	va_start(args, format);
	r = putv(conn, format, NULL, 0, args);
	va_end(args);

	return r;
}

static int paramArrayAssign(
	PGconn *conn,
	char **buf,
	Oid **oids,
	char ***vals,
	int **lens,
	int **fmts)
{
	int n;
	PGparam *param = GETPARAM(conn);

	/* no params to assign */
	if(param->vcnt == 0)
		return 1;

	/* required memory size for the 4 param arrays */
	n = (int)((sizeof(void *) * param->vcnt) + /* values */
			((sizeof(int) * 2) * param->vcnt) +	   /* lengths and formats */
			(sizeof(Oid) * param->vcnt));				   /* oids */

	/* required memory is too large for stack buffer, get some heap */
	if(n > PARAM_STACKSIZE)
	{
		char *p;

		if(!(p = (char *)malloc(n)))
		{
			printfPQExpBuffer(&conn->errorMessage,
				libpq_gettext("cannot allocate param arrays\n"));
			return 0;
		}

		*buf = p;
	}

	/* give arrays memory from buffer, which could be stack or heap. */
	*vals = (char **)*buf;
	*lens = (int *)(*buf + (sizeof(void *) * param->vcnt));
	*fmts = (*lens) + param->vcnt;
	*oids = (Oid *)((*fmts) + param->vcnt);

	/* loop param values and assign value, length, format and oid to arrays. */
	for(n=0; n < param->vcnt; n++)
	{
		(*oids)[n] = param->vals[n].oid;
		(*vals)[n] = param->vals[n].data;
		(*lens)[n] = param->vals[n].datal;
		(*fmts)[n] = param->vals[n].format;
	}

	return 1;
}

static int putptarray(PGconn *conn, int npts, PGpoint *pts, Oid oid, int closed)
{
	int i;
	int hdr = (int)sizeof(int);
	/* 512pts (8k) */
	char _stackbuffer[1 + sizeof(int) + (512 * sizeof(PGpoint))];
	unsigned int *data = (unsigned int *)_stackbuffer;
	unsigned int *d = data;

	/* pts is for a path, include 1 byte open/closed flag */
	if(oid == PATHOID)
		hdr++;

	/* If path is > 512 pts, allocate from the heap */
	if(npts > 512)
	{
		d = data = (unsigned int *)malloc(hdr + (npts * sizeof(PGpoint)));
		if(!d)
			return 0;
	}

	if(oid == PATHOID)
	{
		*(char *)data = closed ? 1 : 0; /* path open/closed flag */
		d = (unsigned int *)(((char *)data) + 1);
	}

	/* write the number of points as an int32 */
	d[0] = (unsigned int)htonl(npts);

	/* assign points to the data 'd' buffer */
	for(i=0, d++; i < npts; i++, d+=4)
	{
		pqSwapi8(d,	    &pts[i].x, 1);
		pqSwapi8(d + 2, &pts[i].y, 1);
	}

	/* put the data buffer, MUST include 'hdr' in the len */
	i = PQput(conn, data, hdr + (npts * sizeof(PGpoint)), oid, 1);

	/* if allocated from the heap, free it */
	if((char *)data != _stackbuffer)
		free(data);

	return i;
}

static int putv(
	PGconn *conn,
	const char *cmdfmt,
	char **out,
	int outl,
	va_list args)
{
	int r=0;
	int count = 0;
	char *cmd = NULL;
	char *cmd_start = NULL;

	if(out)
	{
		/* determine the extra storage needed, which depends on max params */
		if(_PQMAXPARAM > 999)
			r = ((_PQMAXPARAM - 999) * 3) + ((999 - 99) * 2) + 90;
		else if(_PQMAXPARAM > 99)
			r = ((_PQMAXPARAM - 99) * 2) + 90;
		else
			r = _PQMAXPARAM;

		r = strlen(cmdfmt) + r;
		if(r >= outl)
		{
			if(!(*out = (char *)malloc(r+1)))
			{
				printfPQExpBuffer(&conn->errorMessage,
					libpq_gettext("cannot allocate execf format string"));
				return 0;
			}
		}

		cmd_start = cmd = *out;
	}

	while(*cmdfmt != '\0' && count <= _PQMAXPARAM)
	{
		/* skip quoted strings.  Doesn't need to account for E'' syntax. The
		 * E is copied over prior to the quoted string.
		 */
		if(*cmdfmt == '\'')
		{
			const char *end = cmdfmt;

			while(*++end)
			{
				/* If we see a backslash, skip an extra char.  No need to dig any
				 * further since this method works with \digits and \hex.
				 */
				if(*end == '\\')
					end++;
				else if(*end == '\'')
					break;
			}

			if(!*end)
			{
				PGparam *param =  GETPARAM(conn);
				param->vcnt -= count;
				printfPQExpBuffer(&conn->errorMessage,
					libpq_gettext("unterminated single quoted string"));
				return 0;
			}

			end++;

			if(cmd)
			{
				memcpy(cmd, cmdfmt, end - cmdfmt);
				cmd += (end - cmdfmt);
			}

			cmdfmt = end;
			continue;
		}
		else if(*cmdfmt != '%')
		{
			if(cmd)
				*cmd++ = *cmdfmt++;
			continue;
		}

		switch(cmdfmt[1])
		{
			case 'c': /* "char" */
				r = PQputchar(conn, (char)va_arg(args, int));
				break;

			case 'b': /* bool */
				r = PQputbool(conn, va_arg(args, int));
				break;

			case 's': /* short ... int2*/
				r = PQputint2(conn, (short)va_arg(args, int));
				break;

			case 'd': /* int ... int4 */
				r = PQputint4(conn, va_arg(args, int));
				break;

			case 'l': /* long long ... int8 */
			{
				size_t val[2];
#pragma pack(4)
				struct { unsigned int a; unsigned int b; } v;
#pragma pack()

				val[0] = va_arg(args, size_t);
				if(sizeof(size_t) == 4)
					val[1] = va_arg(args, size_t);

				memcpy(&v, val, 8);
				r = PQputint8(conn, v);
				break;
			}

			case 'f': /* float4 */
				r = PQputfloat4(conn, (float)va_arg(args, double));
				break;

			case 'F': /* float8 */
				r = PQputfloat8(conn, va_arg(args, double));
				break;

			case 'a': /* bytea */
			{
				int len = va_arg(args, int);
				void *data = va_arg(args, void *);
				r = PQputbytea(conn, data, len);
				break;
			}

			case 'A': /* byteaptr */
			{
				int len = va_arg(args, int);
				void *data = va_arg(args, void *);
				r = PQputbyteaptr(conn, data, len);
				break;
			}

			case 't': /* text */
				r = PQputtext(conn, va_arg(args, const char *));
				break;

			case 'T': /* textptr */
				r = PQputtextptr(conn, va_arg(args, const char *));
				break;

			case 'p': /* point */
				r = PQputpoint(conn, va_arg(args, PGpoint *));
				break;

			case 'L': /* line segment */
				r = PQputlseg(conn, va_arg(args, PGlseg *));
				break;

			case 'B': /* box */
				r = PQputbox(conn, va_arg(args, PGbox *));
				break;

			case 'C': /* circle */
				r = PQputcircle(conn, va_arg(args, PGcircle *));
				break;

			case 'P': /* path */
				r = PQputpath(conn, va_arg(args, PGpath *));
				break;

			case 'y': /* polygon */
				r = PQputpolygon(conn, va_arg(args, PGpolygon *));
				break;

			case 'i': /* inet & cidr */
				r = PQputinet(conn, va_arg(args, PGinet *));
				break;

			case 'm': /* macaddr */
				r = PQputmacaddr(conn, va_arg(args, PGmacaddr *));
				break;

			case 'S': /* string */
				r = PQputstr(conn, va_arg(args, char *));
				break;

			case 'n': /* NULL */
				r = PQputnull(conn);
				break;

			default:
				r = 0;
				printfPQExpBuffer(&conn->errorMessage,
					libpq_gettext("unknown conversion specifier %.2s"), cmdfmt);
				break;
		}

		if(r == 0)
		{
			PGparam *param =  GETPARAM(conn);
			char buf[512];
			param->vcnt -= count;
			strcpy(buf, conn->errorMessage.data);
			printfPQExpBuffer(&conn->errorMessage,
				libpq_gettext("parameter $%d failed - %s"), count+1, buf);
			return 0;
		}

		cmdfmt += 2;
		count++;

		if(cmd)
			cmd += sprintf(cmd, "$%d", count);
	}

	if(cmd)
		*cmd = 0;

	return r;
}

unsigned int *pqSwapi8(void *outp, void *inp, int tonet)
{
	static int n = 1;
	unsigned int *in = (unsigned int *)inp;
	unsigned int *out = (unsigned int *)outp;

	/* swap when needed */
	if(*(char *)&n == 1)
	{
		out[0] = (unsigned int)(tonet ? htonl(in[1]) : ntohl(in[1]));
		out[1] = (unsigned int)(tonet ? htonl(in[0]) : ntohl(in[0]));
	}
	else
	{
		out[0] = in[0];
		out[1] = in[1];
	}

	return outp;
}



